#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of the various reporter base implementations, plus combinators."""

import itertools

import pytest
import uv.reporter.base as b
import uv.reporter.store as r


@pytest.fixture
def mem():
  return r.MemoryReporter()


class TestReporter(b.AbstractReporter):

  def report(self, step, k, v):
    return None


def test_default_reporter():
  """By default, `reader` isn't implemented."""

  assert TestReporter().reader() is None

  # close does nothing by default.
  assert TestReporter().close() is None


class ThrowCloseReporter(b.AbstractReporter):

  def close(self) -> None:
    raise IOError("Don't close me!")


def test_prefix_reporter(mem):
  prefixed = mem.with_prefix("toots")
  prefixed_reader = prefixed.reader()
  reader = mem.reader()

  # reporting through a prefixed reporter, through either method, should
  # prepend the prefix onto the key before hitting the underlying store.
  prefixed.report_all(0, {"a": 1, "b": 2})
  prefixed.report(1, "a", 2)

  expected = {"toots.a": [1, 2], "toots.b": [2]}

  # if you read through the base reader, you'll need to prepend the prefix.
  assert reader.read_all(["toots.a", "toots.b"]) == expected
  assert prefixed_reader.read_all(["a", "b"]) == expected

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().with_prefix("prefix").close()


def test_prefix_reporter_handles_int_keys(mem):
  prefixed = mem.with_prefix("toots")
  prefixed_reader = prefixed.reader()
  reader = mem.reader()

  prefixed.report_all(0, {0: 1, 1: 2})
  prefixed.report_all(0, {"0": 3})
  prefixed.report(1, 0, 2)

  expected = {"toots.0": [1, 3, 2], "toots.1": [2]}

  # if you read through the base reader, you'll need to prepend the prefix.
  assert reader.read_all(["toots.0", "toots.1"]) == expected
  assert prefixed_reader.read_all([0, 1]) == expected


def test_suffix_reporter(mem):
  suffixed = mem.with_suffix("toots")
  suffixed_reader = suffixed.reader()
  reader = mem.reader()

  # reporting through a suffixed reporter, through either method, should
  # prepend the suffix onto the key before hitting the underlying store.
  suffixed.report_all(0, {"a": 1, "b": 2})
  suffixed.report(1, "a", 2)

  expected = {"a.toots": [1, 2], "b.toots": [2]}

  # if you read through the base reader, you'll need to prepend the suffix.
  assert reader.read_all(["a.toots", "b.toots"]) == expected
  assert suffixed_reader.read_all(["a", "b"]) == expected

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().with_suffix("suffix").close()


class CountCloseReporter(b.AbstractReporter):

  def __init__(self):
    self._closed = 0

  def times_called(self) -> int:
    return self._closed

  def close(self) -> None:
    self._closed += 1


def test_multi_reporter(mem):
  # this compound reporter should 4x add each reported item.
  mem4 = mem.plus(mem, mem, mem)
  reader = mem.reader()

  mem4.report_all(0, {"a": 1})
  mem4.report(1, "b", 2)

  assert reader.read_all(["a", "b"]) == {"a": [1, 1, 1, 1], "b": [2, 2, 2, 2]}

  ccr = CountCloseReporter()
  ccr4 = ccr.plus(ccr, ccr, ccr)
  ccr4.close()

  # the multi-reporter should call the backing store 4 times, since we used
  # copies of the same store.
  assert ccr.times_called() == 4

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().plus(mem).close()


def test_filter_step(mem):
  on_false = r.MemoryReporter()

  even_only = mem.filter_step(lambda step: step % 2 == 0, on_false=on_false)
  reader = even_only.reader()

  even_only.report(0, "a", 0)
  even_only.report(1, "a", 1)
  even_only.report(2, "a", 2)

  even_only.report_all(3, {"b": 1, "c": 2})
  even_only.report_all(4, {"b": 2, "c": 4})

  # only the items written with even steps above should get through.
  assert reader.read_all(["a", "b", "c"]) == {"a": [0, 2], "b": [2], "c": [4]}

  # every time the predicate returned false, items went into the other store.
  assert on_false.reader().read_all(["a", "b", "c"]) == {
      "a": [1],
      "b": [1],
      "c": [2]
  }

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().filter_step(lambda step: True).close()


def test_report_each_n():
  mem1 = r.MemoryReporter().report_each_n(-100)
  mem2 = r.MemoryReporter().report_each_n(1)
  even_only = r.MemoryReporter().report_each_n(2)
  reporter = mem1.plus(mem2, even_only)
  even_reader = even_only.reader()

  reporter.report(0, "a", 0)
  reporter.report(1, "a", 1)
  reporter.report(2, "a", 2)

  reporter.report_all(3, {"b": 1, "c": 2})
  reporter.report_all(4, {"b": 2, "c": 4})

  # only the items written with even steps above should get through.
  assert even_reader.read_all(["a", "b", "c"]) == {
      "a": [0, 2],
      "b": [2],
      "c": [4]
  }

  all_data = {"a": [0, 1, 2], "b": [1, 2], "c": [2, 4]}

  # both mem1 and mem2 receive ALL data.
  assert mem1.reader().read_all(["a", "b", "c"]) == all_data
  assert mem2.reader().read_all(["a", "b", "c"]) == all_data


def test_filter_values(mem):
  on_false = r.MemoryReporter()

  step_plus_v_even = mem.filter_values(lambda step, v: (step + v) % 2 == 0,
                                       on_false=on_false)
  reader = step_plus_v_even.reader()

  step_plus_v_even.report(0, "a", 0)
  step_plus_v_even.report(1, "a", 4)
  step_plus_v_even.report(2, "a", 2)
  step_plus_v_even.report(3, "a", 2)

  step_plus_v_even.report_all(3, {"b": 1, "c": 2})
  step_plus_v_even.report_all(4, {"b": 2, "c": 4})

  # only the items written where the step PLUS the value is even above should
  # get through.
  assert reader.read_all(["a", "b", "c"]) == {
      "a": [0, 2],
      "b": [1, 2],
      "c": [4]
  }

  # every time the predicate returned false, items went into the other store.
  assert on_false.reader().read_all(["a", "b", "c"]) == {
      "a": [4, 2],
      "b": [],
      "c": [2]
  }

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().filter_values(lambda step: True).close()

  # check that the close is passthrough for the on_false reporter:
  with pytest.raises(IOError):
    r.MemoryReporter().filter_values(lambda step: True,
                                     on_false=ThrowCloseReporter()).close()


def test_map_values(mem):
  squared_if_even = mem.map_values(lambda step, v: v * v
                                   if step % 2 == 0 else v)
  reader = mem.reader()

  squared_if_even.report(2, "a", 2)
  squared_if_even.report_all(3, {"b": 1, "c": 7})

  # all of the reported values, when reported through the squared reporter...
  # are squared, IF the step value is even.
  assert reader.read_all(["a", "b", "c", "d"]) == {
      "a": [4],
      "b": [1],
      "c": [7],
      "d": []
  }

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().map_values(lambda step, v: v).close()


def test_stepped(mem):
  stepped_mem = mem.stepped(step_key="this_step")
  reader = mem.reader()

  stepped_mem.report(2, "a", 2)
  stepped_mem.report_all(3, {"b": 1, "c": 7})

  # a stepped reporter converts the incoming data into a dictionary that logs
  # the step as well as the value.
  assert reader.read_all(["a", "b", "c"]) == {
      "a": [{
          "this_step": 2,
          "value": 2
      }],
      "b": [{
          "this_step": 3,
          "value": 1
      }],
      "c": [{
          "this_step": 3,
          "value": 7
      }],
  }


def test_from_thunk(mem):
  counter = itertools.count()

  def thunk():
    return {"counter": next(counter)}

  thunked = mem.from_thunk(thunk)
  reader = thunked.reader()

  # every time thunk(step) gets called, the function above returns a new input
  # to report_all with the supplied step.
  thunked.thunk(0)

  # report and report_all pass through.
  thunked.report(1, "a", 2)
  thunked.thunk(2)
  thunked.report_all(3, {"b": 1, "c": 7})
  thunked.thunk(4)

  assert reader.read_all(["a", "b", "c", "counter"]) == {
      "a": [2],
      "b": [1],
      "c": [7],
      "counter": [0, 1, 2]
  }

  # check that the close is passthrough:
  with pytest.raises(IOError):
    ThrowCloseReporter().from_thunk(lambda: 10).close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base and Combinators; these classes and functions allow you to compose
reporters together into compound reporters.

"""
from abc import ABCMeta
from typing import Any, Callable, Dict, Optional

import uv.reader.base as rb
import uv.types as t
import uv.util.attachment as a


class AbstractReporter(metaclass=ABCMeta):
  """Base class for all reporters. A reporter is a type that is able to log
  timeseries of values for different t.MetricKey instances, one item at a time.

  NOTE - by default, report_all and report are implemented in terms of one
  another. This means that you can choose which method you'd like to override,
  or override both... but if you don't override any you'll see infinite
  recursion.

  Be careful not to abuse the kindness!

  """

  def report_param(self, k: str, v: str) -> None:
    """Accepts a key and value parameter and logs these as parameters alongside the
    reported metrics.

    """
    return None

  def report_params(self, m: Dict[str, str]) -> None:
    """Accepts a dict of parameter name -> value, and logs these as parameters
    alongside the reported metrics.

    """
    return None

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    """Accepts a step (an ordered int referencing some timestep) and a dictionary
    of metric key => metric value, and persists the metric into some underlying
    store.

    Extending classes are expected to perform some side effect that's either
    visually useful, as in a live-plot, or recoverable via some matching
    extension of AbstractReader.

    """

    for k, v in m.items():
      self.report(step, k, v)

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    """Accepts a step (an ordered int referencing some timestep), a metric key and
    a value, and persists the metric into some underlying store.

    """
    return self.report_all(step, {k: v})

  def reader(self) -> Optional[rb.AbstractReader]:
    """Returns an implementation of AbstractReader that can access the data in this
    store.

    Returns None by default; extending classes are encouraged, but not
    required, to override.

    """
    return None

  def close(self) -> None:
    """Release any resources held open by this reporter instance."""
    return None

  # Combinators.

  def with_prefix(self, prefix: t.Prefix):
    """Returns an instance of PrefixedReporter wrapping the current instance. This
    reporter attaches the supplied prefix to every metric key it sees before
    passing metrics on.

    """
    return PrefixedReporter(self, prefix)

  def with_suffix(self, suffix: t.Suffix):
    """Returns an instance of SuffixedReporter wrapping the current instance. This
    reporter attaches the supplied suffix to every metric key it sees before
    passing metrics on.

    """
    return SuffixedReporter(self, suffix)

  def plus(self, *others: "AbstractReporter"):
    """Returns an instance of MultiReporter wrapping the current instance. This
    reporter broadcasts its inputs to this instance, plus any other reporters
    supplied to this method, every time it sees a metric passed in via report
    or report_all.

    """
    return MultiReporter(self, *others)

  def filter_step(self,
                  pred: Callable[[int], bool],
                  on_false: Optional["AbstractReporter"] = None):
    """Accepts a predicate function from step to boolean, and returns a reporter
    that tests every step against the supplied function. If the function
    returns true, metrics get passed on to this reporter; else, they get
    filtered out.

    If a reporter is supplied to on_false, any time the predicate returns false
    items are routes to that store instead of base.

    """

    def step_pred(step, _):
      return pred(step)

    return FilterValuesReporter(self, step_pred, on_false_reporter=on_false)

  def filter_values(self,
                    pred: Callable[[int, t.Metric], bool],
                    on_false: Optional["AbstractReporter"] = None):
    """"Accepts a function from (step, metric) to boolean; every (step, metric)
    pair passed to report and report_all are passed into this function. If the
    predicate returns true, the metric is passed on; else, it's filtered.

    """
    return FilterValuesReporter(self, pred, on_false_reporter=on_false)

  def map_values(self, fn: Callable[[int, t.Metric], t.Metric]):
    """"Accepts a function from (step, metric) to some new metric; every (step,
    metric) pair passed to report and report_all are passed into this function,
    and the result is passed down the chain to this, the calling reporter.

    """
    return MapValuesReporter(self, fn)

  def stepped(self, step_key: Optional[str] = None):
    """Returns a new reporter that modifies incoming metrics by wrapping them in a
    dict of this form before passing them down to this instance of reporter:

    {step_key: step, "value": metric_value}

    where step_key is the supplied argument, and equal to "step" by default.
    This is useful for keeping track of each metric's timestamp.

    """
    return stepped_reporter(self, step_key=step_key)

  def report_each_n(self, n: int):
    """Returns a new reporter that only reports every n steps; specifically, the
    new reporter will only accept metrics where step % n == 0.

    If n <= 1, this reporter, untouched, is returned directly.

    """
    n = max(1, n)
    if n > 1:
      return self.filter_step(lambda step: step % n == 0)
    else:
      return self

  def from_thunk(self, thunk: Callable[[], Dict[t.MetricKey, t.Metric]]):
    """Returns a new Reporter that passes all AbstractReporter methods through, but
    adds a new method called "thunk()" that, when called, will pass the emitted
    map of metric key to metric down to the underlying store.

    thunk() returns the value emitted by the no-arg function passed here via
    `thunk`.
    """
    return ThunkReporter(self, thunk)


# Combinators


def stepped_reporter(base: AbstractReporter,
                     step_key: Optional[str] = None) -> AbstractReporter:
  """Returns a new reporter that modifies incoming metric by wrapping them in a
  dict of this form before passing them down to base:

    {step_key: step, "value": metric_value}

  where step_key is the supplied argument, and equal to "step" by default. This
  is useful for keeping track of each metric's timestamp.

  """
  if step_key is None:
    step_key = "step"

  def _augment(step: int, v: Any) -> Dict[str, Any]:
    return {step_key: step, "value": v}

  return MapValuesReporter(base, _augment)


class FilterValuesReporter(AbstractReporter):
  """Reporter that filters incoming metrics by applying a predicate from (step,
  t.Metric). If true, the reporter passes the result on to the underlying
  reporter. Else, nothing.

  Args:
    base: Backing reporter. All report and report_all calls proxy here.
    predicate: function from (step, metric) to metric.

  """

  def __init__(self,
               base: AbstractReporter,
               predicate: Callable[[int, t.Metric], bool],
               on_false_reporter: Optional[AbstractReporter] = None):
    self._base = base
    self._pred = predicate
    self._on_false_reporter = on_false_reporter

  def report_param(self, k: str, v: str) -> None:
    return self._base.report_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    return self._base.report_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    good = {k: v for k, v in m.items() if self._pred(step, v)}
    bad = {k: v for k, v in m.items() if not self._pred(step, v)}

    if good:
      self._base.report_all(step, good)

    if self._on_false_reporter and bad:
      self._on_false_reporter.report_all(step, bad)

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    if self._pred(step, v):
      self._base.report(step, k, v)

    elif self._on_false_reporter:
      self._on_false_reporter.report(step, k, v)

  def reader(self) -> Optional[rb.AbstractReader]:
    return self._base.reader()

  def close(self) -> None:
    self._base.close()

    if self._on_false_reporter:
      self._on_false_reporter.close()


class MapValuesReporter(AbstractReporter):
  """Reporter that modifies incoming metrics by applying a function from (step,
  t.Metric) to a new metric before passing the result on to the underlying
  reporter.

  Args:
    base: Backing reporter. All report and report_all calls proxy here.
    fn: function from (step, metric) to metric.

  """

  def __init__(self, base: AbstractReporter, fn: Callable[[int, t.Metric],
                                                          t.Metric]):
    self._base = base
    self._fn = fn

  def report_param(self, k: str, v: str) -> None:
    return self._base.report_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    return self._base.report_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    self._base.report_all(step, {k: self._fn(step, v) for k, v in m.items()})

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    self._base.report(step, k, self._fn(step, v))

  def reader(self) -> Optional[rb.AbstractReader]:
    return self._base.reader()

  def close(self) -> None:
    self._base.close()


class MultiReporter(AbstractReporter):
  """Reporter that broadcasts out metrics to all N reporters supplied to its
  constructor.

  Args:
    reporters: instances of t.AbstractReporter that will receive all calls to
               this instance's methods.

  """

  def __init__(self, *reporters: AbstractReporter):
    self._reporters = reporters

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    for r in self._reporters:
      r.report_all(step, m)

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    for r in self._reporters:
      r.report(step, k, v)

  def close(self) -> None:
    for r in self._reporters:
      r.close()


class PrefixedReporter(AbstractReporter):
  """Reporter that prepends a prefix to all keys before passing requests on to
  the supplied backing reporter.

  Args:
    base: Backing reporter. All report and report_all calls proxy here.
    prefix: the prefix to attach to all keys supplied to any method.

  """

  def __init__(self, base: AbstractReporter, prefix: t.Prefix):
    self._base = base
    self._prefix = prefix

  def report_param(self, k: str, v: str) -> None:
    return self._base.report_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    return self._base.report_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    self._base.report_all(step, a.attach(m, self._prefix, prefix=True))

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    newk = a.attach_s(k, self._prefix, prefix=True)
    self._base.report(step, newk, v)

  def reader(self) -> Optional[rb.AbstractReader]:
    return rb.PrefixedReader(self._base.reader(), self._prefix)

  def close(self) -> None:
    self._base.close()


class SuffixedReporter(AbstractReporter):
  """Reporter that prepends a prefix to all keys before passing requests on to
  the supplied backing reporter.

  Args:
    base: Backing reporter. All report and report_all calls proxy here.
    suffix: the suffix to attach to all keys supplied to any method.

  """

  def __init__(self, base: AbstractReporter, suffix: t.Suffix):
    self._base = base
    self._suffix = suffix

  def report_param(self, k: str, v: str) -> None:
    return self._base.report_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    return self._base.report_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    self._base.report_all(step, a.attach(m, self._suffix, prefix=False))

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    newk = a.attach_s(k, self._suffix, prefix=False)
    self._base.report(step, newk, v)

  def reader(self) -> Optional[rb.AbstractReader]:
    return rb.SuffixedReader(self._base.reader(), self._suffix)

  def close(self) -> None:
    self._base.close()


class ThunkReporter(AbstractReporter):
  """Reporter that passes all AbstractReporter methods through, but adds a new
  method called "thunk()" that, when called, will pass the emitted map of
  metric key to metric down to the underlying store.

  Args:
    base: Backing reporter. All report and report_all calls proxy here.
    thunk: no-arg lambda that returns a metric dictionary.

  """

  def __init__(self, base: AbstractReporter, thunk):
    self._base = base
    self._thunk = thunk

  def thunk(self, step: int) -> None:
    self.report_all(step, self._thunk())

  def report_param(self, k: str, v: str) -> None:
    return self._base.report_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    return self._base.report_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    self._base.report_all(step, m)

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    self._base.report(step, k, v)

  def reader(self) -> Optional[rb.AbstractReader]:
    return self._base.reader()

  def close(self) -> None:
    self._base.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from uv.reporter.base import AbstractReporter
from uv.reporter.state import (active_reporter, get_reporter, report,
                               report_all, report_param, report_params,
                               set_reporter, start_run)
from uv.reporter.store import LoggingReporter, MemoryReporter

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

__all__ = [
    "active_reporter",
    "get_reporter",
    "set_reporter",
    "report",
    "report_all",
    "report_param",
    "report_params",
    "start_run",
    "AbstractReporter",
    "LoggingReporter",
    "MemoryReporter",
]



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities shared by functions that interact with pyfilesystem2."""

import io
from typing import Any, Union

import fs as pyfs
import uv.types as t
import uv.util as u
from casfs import CASFS
from fs.base import FS
from fs.copy import copy_fs
from fs.zipfs import ZipFS


class HandleCache():
  """Class that opens file-handles and maintains an internal cache of those
  handles. On close, closes the supplied filesystem, closes all handles and
  clears the cache.

  """

  def __init__(self, fs: FS):
    self._m = {}
    self._fs = fs

  def open(self, path: str, mode: str):
    abs_path = pyfs.path.abspath(path)
    handle = self._m.get(abs_path)

    if handle is None:
      handle = self._fs.open(abs_path, mode=mode)
      self._m[abs_path] = handle

    return handle

  def clear(self) -> None:
    for _, handle in self._m.items():
      handle.close()

    self._m.clear()

  def close(self) -> None:
    self.clear()
    self._fs.close()


def load_fs(root: Union[FS, str]) -> FS:
  """If str is supplied, returns an instance of OSFS, backed by the local
    filesystem; else, returns the filesystem argument directly. Errors if
    supplied with an invalid argument.

    """
  if isinstance(root, str):
    return pyfs.open_fs(root, create=True)

  if isinstance(root, FS):
    return root

  raise ValueError("Not a filesystem or path!")


def jsonl_path(k: t.MetricKey) -> str:
  """Returns an absolute path with an appropriate suffix for a jsonl file."""
  return pyfs.path.abspath("{}.jsonl".format(k))


def to_bytes(item: Union[str, bytes]) -> bytes:
  """Accepts either a bytes instance or a string; if str, returns a bytes
  instance, else acts as identity.

  """
  ret = item

  if not isinstance(item, bytes):
    ret = bytes(item, "utf8")

  return ret


def jsonl_bytes(v: Any) -> bytes:
  """Returns a bytes instance containing a single line of json with a newline
  appended.

  """
  return to_bytes(u.json_str(v) + "\n")


def get_cas(cas_input):
  """Version of the CASFS constructor that creates directories that don't exist.
  TODO delete this once we get that idea merged back into CASFS.

  """
  cas_fs = pyfs.open_fs(cas_input, create=True)
  return CASFS(cas_fs)


def persist_to_cas_via_memory(cfs, source_fs):
  """Example call:

  persist_to_cas_via_memory(
    CASFS("/Users/samritchie/casfs"),
    "/Users/samritchie/tester" # metrics directory
  )

  """
  with io.BytesIO() as f:
    # make sure the ZipFS closes before we attempt to transfer its contents
    # over to the content addressable store.
    with ZipFS(f, write=True) as zfs:
      copy_fs(source_fs, zfs)

    return cfs.put(f)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AbstractReporter implementations that live at the bottom of the reporter
stack. These reporters aren't combinators; they're responsible for persisting
metrics into underlying store or mechanism.

"""

import sys
from typing import Callable, Dict, List, Optional

import uv.reader.base as rb
import uv.reader.store as rs
import uv.types as t
import uv.util as u
from uv.reporter.base import AbstractReporter

# Class Definitions


class NullReporter(AbstractReporter):
  """Reporter that does nothing with the metrics passed to its various methods.
  reader() returns an instance of rs.EmptyReader.

  """

  def report_all(self, step, m):
    return None

  def report(self, step, k, v):
    return None

  def reader(self) -> Optional[rb.AbstractReader]:
    return rs.EmptyReader()


class LambdaReporter(AbstractReporter):
  """AbstractReporter implementation that defers to a supplied lambda for its
  persistence ability. This allows you to escape the object-oriented
  programming paradigm, if you so choose.

  Args:
    report: Function called whenever reporter.report(step, k, v) is called.
    report_all: Function called whenever reporter.report_all(step, m) is
                called.
    report_param: Function called whenever reporter.report_param(k, v) is
                  called.
    report_params: Function called whenever reporter.report_params(m) is
                   called.
    close: If supplied, this no-arg function will get called by this instance's
           `close` method.

  """

  def __init__(self,
               report: Optional[Callable[[int, t.MetricKey, t.Metric],
                                         None]] = None,
               report_all: Optional[Callable[[int, Dict[t.MetricKey, t.Metric]],
                                             None]] = None,
               report_param: Optional[Callable[[str, str], None]] = None,
               report_params: Optional[Callable[[Dict[str, str]], None]] = None,
               close: Optional[Callable[[], None]] = None):
    if report is None and report_all is None:
      raise ValueError(
          "Must supply one of `report` and `report_all` to `LambdaReporter`.")

    self._reportparam = report_param
    self._reportparams = report_params
    self._report = report
    self._reportall = report_all
    self._close = close

  def report_param(self, k: str, v: str) -> None:
    if self._reportparam is None:
      super().report_param(k, v)
    else:
      self._reportparam(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    if self._reportparams is None:
      super().report_params(m)
    else:
      self._reportparams(m)

  def report_all(self, step, m):
    if self._reportall is None:
      super().report_all(step, m)
    else:
      self._reportall(step, m)

  def report(self, step, k, v):
    if self._report is None:
      super().report(step, k, v)
    else:
      self._report(step, k, v)

  def close(self) -> None:
    if self._close is not None:
      self._close()


class LoggingReporter(AbstractReporter):
  """Reporter that logs all data to the file handle you pass in using a fairly
  sane format. Compatible with tqdm, the python progress bar.

  """

  @staticmethod
  def tqdm():  # pragma: no cover
    """Returns a logging reporter that will work with a tqdm progress bar."""
    return LoggingReporter(u.TqdmFile(sys.stderr))

  def __init__(self, file=sys.stdout, digits: int = 3):

    self._file = file
    self._digits = digits

  def _format(self, v: t.Metric) -> str:
    """Formats the value into something appropriate for logging."""
    if u.is_number(v):
      return "{num:.{digits}f}".format(num=float(v), digits=self._digits)

    return str(v)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    s = ", ".join(["{} = {}".format(k, self._format(v)) for k, v in m.items()])
    f = self._file
    print("Step {}: {}".format(step, s), file=f)


class MemoryReporter(AbstractReporter):
  """Reporter that stores metrics in a Python dictionary, keyed by t.MetricKey.
  Metrics are stored as a list.

  Args:
    m: Optional dictionary mapping metric keys to a list of accumulated metric
       values. If supplied, this dictionary will be mutated as new metrics
       arrive.

  """

  def __init__(self,
               m: Optional[Dict[str, List[t.Metric]]] = None,
               params_store: Optional[Dict[str, str]] = None):
    if m is None:
      m = {}

    if params_store is None:
      params_store = {}

    self._m = m
    self._params = params_store

  def report_param(self, k: str, v: str) -> None:
    self._params[k] = v

  def report_params(self, m: Dict[str, str]) -> None:
    self._params.update({**m})

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    for k, v in m.items():
      self._m.setdefault(str(k), []).append(v)

  def clear(self):
    """Erase all key-value pairs in the backing store."""
    self._m.clear()

  def reader(self) -> Optional[rb.AbstractReader]:
    return rs.MemoryReader(self._m)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import uv.util.env as ue


def test_extract_params(monkeypatch):

  def mem_env(prefix):
    return {
        f"{prefix}_MY_KEY": "face",
        f"{prefix}_ANOTHER_KEY": "sandwich",
        f"{prefix}THIRD_KEY": "ham"
    }

  expected = {"my_key": "face", "another_key": "sandwich", "third_key": "ham"}

  # with various prefixes, a custom-supplied environment will return the
  # correctly parsed env variables.
  assert expected == ue.extract_params(prefix="ENVVAR", env=mem_env("ENVVAR"))
  assert expected == ue.extract_params(prefix="funky", env=mem_env("funky"))

  k = f"{ue._ENV_VAR_PREFIX}_RANDOM_KEY"
  v = "better_not_be_set"

  # make sure we don't have some random value set
  if os.environ.get(k):
    monkeypatch.delenv(k)

  # the environment should be empty.
  assert ue.extract_params() == {}

  # set our expected kv pair...
  monkeypatch.setenv(k, v)

  # and get it back from the env.
  assert ue.extract_params() == {"random_key": v}



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The base AbstractReader class, along with all of the subclasses that
AbstractReader is able to return from any of its methods."""

from abc import ABC, ABCMeta, abstractmethod
from typing import Dict, Iterable, List

import uv.types as t
import uv.util.attachment as a


class IterableReader(ABC):
  """Class marker for Reader instances that are able to provide an iterable of
  all keys that it's possible to read using this Reader instance.

  """

  @abstractmethod
  def keys(self) -> Iterable[t.MetricKey]:
    """Returns an iterable of all keys that you're able to query from this
    store.

    """


class AbstractReader(metaclass=ABCMeta):
  """Base class for all classes that are able to track and return lists of
  metrics, keyed by a supplied t.MetricKey.

  NOTE - by default, read_all and read are implemented in terms of one another.
  This means that you can choose which method you'd like to override, or
  override both... but if you don't override any you'll see infinite recursion.

  Be careful not to abuse the kindness!

  """

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    """Accepts a list of t.MetricKey instances and returns a mapping of t.MetricKey
    to a list of all metrics ever logged using that MetricKey.

    The interface requires that the set of keys in the returned dictionary
    equal the input set; any missing key in the store has to map to a default
    empty list.

    """
    return {k: self.read(k) for k in ks}

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    """Returns a list of all metrics ever logged for the supplied t.MetricKey, or
    an empty list if the backing store has no record.

    """
    return self.read_all([k])[k]

  def close(self) -> None:
    """Release any resources held open by this reader instance."""
    return None

  # Combinators

  def with_prefix(self, prefix: t.Prefix):
    """"Returns a new reader that will pass through all methods to this reader
    instance; the only difference will be that every t.MetricKey instance
    supplied to read or read_all will have the supplied prefix attached.

    """
    return PrefixedReader(self, prefix)

  def with_suffix(self, suffix: t.Suffix):
    """"Returns a new reader that will pass through all methods to this reader
    instance; the only difference will be that every t.MetricKey instance
    supplied to read or read_all will have the supplied suffix attached.

    """
    return SuffixedReader(self, suffix)


class PrefixedReader(AbstractReader):
  """Reader that prepends a prefix to all keys before passing requests on to the
  supplied backing store.

  Args:
    base: Backing reader. All read and read_all calls proxy here.
    prefix: the prefix to attach to all keys supplied to any method.

  """

  def __init__(self, base: AbstractReader, prefix: t.Prefix):
    self._base = base
    self._prefix = prefix

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    return self._base.read_all(a.attach_iter(ks, self._prefix, prefix=True))

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    return self._base.read(a.attach_s(k, self._prefix, prefix=True))

  def close(self) -> None:
    self._base.close()


class SuffixedReader(AbstractReader):
  """Reader that prepends a suffix to all keys before passing requests on to the
  supplied backing store.

  Args:
    base: Backing reader. All read and read_all calls proxy here.
    suffix: the suffix to attach to all keys supplied to any method.

  """

  def __init__(self, base: AbstractReader, suffix: t.Suffix):
    self._base = base
    self._suffix = suffix

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    return self._base.read_all(a.attach_iter(ks, self._suffix, prefix=False))

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    return self._base.read(a.attach_s(k, self._suffix, prefix=False))

  def close(self) -> None:
    self._base.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Tensorboard reporter."""

import os
from contextlib import closing
from typing import Any, Dict, Iterable

from tensorflow.core.util import event_pb2
from tensorflow.python.framework import tensor_util
from tensorflow.python.lib.io import tf_record
from tensorflow.python.platform import gfile

import uv.tensorflow.reporter as r


def events_from_file(filepath: str):
  """Returns all events in a single event file.

  Args:
    filepath: Path to the event file.
  Returns:
    A list of all tf.Event protos in the event file.
  """
  records = list(tf_record.tf_record_iterator(filepath))
  result = []
  for rec in records:
    event = event_pb2.Event()
    event.ParseFromString(rec)
    result.append(event)
  return result


def events_from_logdir(log_dir: str):
  """Returns all events in the single eventfile in logdir.

  Args:
    logdir: The directory in which the single event file is sought.
  Returns:
    A list of all tf.Event protos from the single event file.
  Raises:
    AssertionError: If logdir does not contain exactly one file.

  """
  assert gfile.Exists(log_dir)
  files = gfile.ListDirectory(log_dir)
  assert len(files) == 1, 'Found not exactly one file in logdir: %s' % files
  return events_from_file(os.path.join(log_dir, files[0]))


def to_numpy(summary_value):
  """Converts the value of the supplied summary to a numpy array."""
  return tensor_util.MakeNdarray(summary_value.tensor)


def is_summary(event: event_pb2.Event) -> bool:
  """Returns true if the supplied proto event represents a summary that we've
logged, false otherwise."""
  return len(event.summary.value) > 0


def get_summaries(
    events: Iterable[event_pb2.Event]) -> Iterable[Dict[str, Any]]:
  """Returns a sequence of dictionaries represen"""

  def process(e):
    v = e.summary.value[0]
    return {"step": e.step, "k": v.tag, "v": to_numpy(v)}

  return [process(e) for e in events if is_summary(e)]


def test_tensorboard_reporter(tmpdir):
  dir_name = str(tmpdir)

  # report all events.
  with closing(r.TensorboardReporter(dir_name)) as reporter:
    reporter.report_all(0, {"a": 1, "b": 10})
    reporter.report_all(1, {"a": 2, "b": 10, "c": 3})
    reporter.report_all(2, {"a": 3, "b": 11})
    reporter.report(3, "a", 8)
    reporter.report(3, "b", 2)
    reporter.report(6, "b", 10)

  # get all the reported events back out:
  events = events_from_logdir(dir_name)
  ms = get_summaries(events)

  # check that all events were written.
  assert len(ms) == 10

  # check all of the A values:
  assert list(filter(lambda m: m["k"] == "a", ms)) == [{
      "step": 0,
      "k": "a",
      "v": 1
  }, {
      "step": 1,
      "k": "a",
      "v": 2
  }, {
      "step": 2,
      "k": "a",
      "v": 3
  }, {
      "step": 3,
      "k": "a",
      "v": 8
  }]

  # Check all of the b values:
  assert list(filter(lambda m: m["k"] == "b", ms)) == [{
      "step": 0,
      "k": "b",
      "v": 10
  }, {
      "step": 1,
      "k": "b",
      "v": 10
  }, {
      "step": 2,
      "k": "b",
      "v": 11
  }, {
      "step": 3,
      "k": "b",
      "v": 2
  }, {
      "step": 6,
      "k": "b",
      "v": 10
  }]

  # specifically check the C value:
  assert list(filter(lambda m: m["k"] == "c", ms)) == [{
      "step": 1,
      "k": "c",
      "v": 3
  }]



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow / Tensorboard reporter that conforms to UV's reporter interface.

In its current form, the TensorboardReporter is able to report scalars.

TODO Add support for logging images directly using:
https://github.com/tensorflow/tensorflow/blob/r2.1/tensorflow/python/keras/callbacks.py#L1400

TODO investigate docs on how to serialize entries:
https://github.com/tensorflow/tensorboard/blob/master/docs/r1/summaries.md

"""

from typing import Dict, Optional

import tensorflow as tf

import uv.reporter.base as b
import uv.types as t


class TensorboardReporter(b.AbstractReporter):
  """Reporter implementation that accepts scalar metrics and reports them to an
  underlying Tensorboard file writer in the supplied directory.

  TensorboardReporter's log_dir argument can be either a local path or a GCloud
  path, prefixed with gs://.

  Arguments to the TensorboardReporter constructor match the arguments of
  https://www.tensorflow.org/api_docs/python/tf/summary/create_file_writer.



  Args:
    log_dir: a string specifying the directory in which to write an event file.
    max_queue: the largest number of summaries to keep in a queue; will
     flush once the queue gets bigger than this. Defaults to 100.
    flush_millis: the largest interval between flushes. Defaults to 120,000.
    filename_suffix: optional suffix for the event file name. Defaults to `.v2`.
    name: a name for the op that creates the writer.

  """

  def __init__(
      self,
      log_dir: str,
      max_queue: Optional[int] = 100,
      flush_millis: Optional[int] = None,
      filename_suffix: Optional[str] = None,
  ):
    self._path = log_dir
    self._writer = tf.summary.create_file_writer(
        self._path,
        max_queue=max_queue,
        flush_millis=flush_millis,
        filename_suffix=filename_suffix)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    with self._writer.as_default():
      for k, v in m.items():
        tf.summary.scalar(k, v, step=step)

    self._writer.flush()

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    with self._writer.as_default():
      tf.summary.scalar(k, v, step=step)

  def close(self) -> None:
    self._writer.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of the base reader and combinators."""

from typing import Dict, List

import hypothesis.strategies as st
from hypothesis import given

import pytest
import uv.reader.base as b
import uv.reader.store as rs
import uv.types as t
import uv.util.attachment as a


class JustRead(b.AbstractReader):
  """Class that ONLY overrides read, not read_all."""

  def __init__(self, base: b.AbstractReader):
    self._base = base

  def read(self, k: str) -> List[t.Metric]:
    return self._base.read(k)


class JustReadAll(b.AbstractReader):
  """Class that ONLY overrides read_all, not read."""

  def __init__(self, base: b.AbstractReader):
    self._base = base

  def read_all(self, ks: List[str]) -> Dict[str, List[t.Metric]]:
    return self._base.read_all(ks)


@given(
    st.dictionaries(st.text(min_size=1),
                    st.lists(st.integers(), max_size=100),
                    max_size=100))
def test_default_read_implementations(m):
  """Check that the read and read_all implementation that, by default, call each
  other, actually work.

  """
  mem = rs.MemoryReader(m)
  just_read = JustRead(mem)
  just_read_all = JustReadAll(mem)

  ks = m.keys()
  just_read_results = just_read.read_all(ks)
  assert just_read_results == just_read_all.read_all(ks)
  assert just_read_results == m

  for k in ks:
    just_read_results = just_read.read(k)
    assert just_read_results == just_read_all.read(k)
    assert just_read_results == m[k]

  assert just_read.close() is None


@given(st.text(min_size=1),
       st.dictionaries(st.text(min_size=1),
                       st.lists(st.integers(), max_size=100),
                       max_size=100))
def test_prefixed_reader(prefix, m):
  """Test that prefixes work."""
  prefixed = {a.attach_s(k, prefix, prefix=True): v for k, v in m.items()}
  mem = rs.MemoryReader(prefixed)

  # the bare store, when queried with keys with the prefix attached, returns
  # the right stuff.
  assert mem.read_all(prefixed.keys()) == prefixed

  reader = mem.with_prefix(prefix)
  # If instead you generate a new reader using with_prefix, you don't need to
  # pass prefixed keys in to read_all; they'll get automatically attached.
  assert reader.read_all(m.keys()) == prefixed

  # read implementation works too.
  for k, v in m.items():
    assert reader.read(k) == v


@given(st.text(min_size=1),
       st.dictionaries(st.text(min_size=1),
                       st.lists(st.integers(), max_size=100),
                       max_size=100))
def test_suffixed_reader(suffix, m):
  """Test that suffixes work."""
  suffixed = {a.attach_s(k, suffix, prefix=False): v for k, v in m.items()}
  mem = rs.MemoryReader(suffixed)

  # the bare store, when queried with keys with the suffix attached, returns
  # the right stuff.
  assert mem.read_all(suffixed.keys()) == suffixed

  reader = mem.with_suffix(suffix)
  # If instead you generate a new reader using with_suffix, you don't need to
  # pass suffixed keys in to read_all; they'll get automatically attached.
  assert reader.read_all(m.keys()) == suffixed

  # read implementation works too.
  for k, v in m.items():
    assert reader.read(k) == v


class ThrowCloseReader(b.AbstractReader):

  def close(self) -> None:
    raise IOError("Don't close me!")


def check_passthrough_close(base_to_reader_fn):
  """Takes a function from a base to reader, and checks that calling close on the
  returned reader actually passes through to the base.

  """
  with pytest.raises(IOError):
    base = ThrowCloseReader()
    base_to_reader_fn(base).close()


def test_attached_reader_passthrough_close():
  check_passthrough_close(lambda s: b.PrefixedReader(s, prefix='face'))
  check_passthrough_close(lambda s: b.SuffixedReader(s, suffix='face'))



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the prefix utilities."""

import hypothesis.strategies as st
from hypothesis import given

import uv.util.attachment as a


@given(st.dictionaries(st.text(min_size=1), st.integers()))
def test_by_prefix(m):
  wrapped = a.by_prefix({"test": m, "train": m})

  for k, vs in m.items():
    assert wrapped.get(a.attach_s(k, "test")) == vs
    assert wrapped.get(a.attach_s(k, "train")) == vs
    assert wrapped.get(a.attach_s(k, "random")) is None


@given(st.dictionaries(st.text(min_size=1), st.integers()))
def test_by_suffix(m):
  wrapped = a.by_suffix({"test": m, "train": m})

  for k, vs in m.items():
    assert wrapped.get(a.attach_s(k, "test", prefix=False)) == vs
    assert wrapped.get(a.attach_s(k, "train", prefix=False)) == vs
    assert wrapped.get(a.attach_s(k, "random", prefix=False)) is None



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reader implementation that relies on the PyFilesystem2 abstraction for
filesystems. See https://github.com/PyFilesystem/pyfilesystem2 for more
information on the library.

"""

import json
from typing import Iterable, List, Union

import fs as pyfs
import uv.fs.util as u
import uv.types as t
from casfs.base import CASFS, Key
from fs.base import FS
from fs.zipfs import ZipFS
from uv.reader.base import AbstractReader, IterableReader


class FSReader(AbstractReader, IterableReader):
  """AbstractReader implementation backed by an instance of pyfilesystem2's FS
  abstraction.

  Args:
    fs: Either an fs URI string, or an actual fs.base.FS object.

  """

  def __init__(self, fs: Union[FS, str]):
    self._fs = u.load_fs(fs)

  def keys(self) -> Iterable[t.Metric]:
    """Returns all files in the filesystem that plausibly contains metrics in jsonl
    format.

    """
    for p in self._fs.walk.files(filter=['*.jsonl']):
      base = pyfs.path.basename(p)
      k, _ = pyfs.path.splitext(base)
      yield k

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    try:
      abs_path = u.jsonl_path(k)
      with self._fs.open(abs_path, mode='rb') as handle:
        lines = handle.read().splitlines()
        return [json.loads(s.decode("utf-8")) for s in lines]

    except pyfs.errors.ResourceNotFound:
      return []

  def close(self) -> None:
    self._fs.close()


class CASReader(FSReader):
  """Override that closes the file too."""

  def __init__(self, cas: CASFS, k: Key):
    self._handle = cas.open(k)
    zfs = ZipFS(self._handle)

    super(CASReader, self).__init__(zfs)

  def close(self) -> None:
    super(CASReader, self).close()
    self._handle.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the functions that live in the __init__.py file."""

import json

import hypothesis.strategies as st
import numpy as np
import tensorflow as tf
from hypothesis import given

import pytest
import uv.util as u


class Wrapper():
  """Wrapper class that has a numpy method, for testing."""

  def __init__(self, x):
    self._x = x

  def numpy(self):
    """Using this instead of importing tensorflow..."""
    return np.float64(self._x)


def test_to_metric():
  x = Wrapper(100)
  assert u.to_metric(100) == 100
  assert u.to_metric(x) == np.float64(100)


def test_to_serializable():
  """Check that various non-serializable things can in fact be serialized using
  this dispatch method."""
  f = 100.0

  # json can't serialize float32:
  with pytest.raises(TypeError):
    json.dumps(np.float32(f))

  # but it can here, if we pass it through to_serializable.
  assert json.dumps(u.to_serializable(np.float32(f))) == str(f)

  # this passthrough automatically using u.json_str.
  assert u.json_str(f) == str(f)

  # by default, to make something serializable, turn it into a string.
  assert u.to_serializable("face") == "face"

  # check that numpy arrays serialize too.
  assert u.to_serializable(np.zeros(shape=(2, 2))) == [[0.0, 0.0], [0.0, 0.0]]


def test_tensor_serialization():
  # check that an int32-tensorflow constant can get to something serializable.
  x = tf.constant(10)
  assert u.to_serializable(x) == 10

  # the serialization passes through, without getting turned into a string:
  assert u.json_str({"x": x}) == '{"x": 10}'

  # floats pass through as well.
  assert u.to_serializable(tf.constant(10.5)) == 10.5


def test_wrap():
  """Test the best-effort wrapping of non-lists."""

  assert u.wrap([1]) == [1]
  assert u.wrap(1) == [1]
  assert u.wrap(None) == []
  assert u.wrap((1, 2)) == [(1, 2)]


def test_is_number():
  """This is meant to check if something can be emitted as a number via json."""
  assert u.is_number("100")
  assert u.is_number(np.float32(100))
  assert u.is_number(np.float64(100))
  assert not u.is_number("face")


class MemFile():
  """Dummy file-like that will write to a list."""

  def __init__(self):
    self._l = []

  def items(self):
    return self._l

  def write(self, x):
    self._l.append(x)

  def flush(self):
    raise IOError("NO FLUSH!")

  def isatty(self):
    raise IOError("NO ATTY!")


@given(st.lists(st.text(min_size=1)))
def test_tqdm_file(lines):
  """Test properties of the TQDM logger - check that it properly passes through
  Tqdm.

  """
  mem = MemFile()
  f = u.TqdmFile(file=mem)

  for l in lines:
    f.write(l)

  def strip(items):
    return [l for l in items if len(l.rstrip()) > 0]

  assert strip(mem.items()) == strip(lines)

  # test passthroughs:
  with pytest.raises(IOError):
    f.flush()

  with pytest.raises(IOError):
    f.isatty()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AbstractReader implementations that live at the bottom of the reader stack.
These readers aren't combinators; they're responsible for returning stored
values from some underlying store or mechanism.

"""

from typing import Callable, Dict, Iterable, List, Optional

import uv.types as t
from uv.reader.base import AbstractReader, IterableReader


class EmptyReader(AbstractReader, IterableReader):
  """AbstractReader that uses NO internal state. EmptyReader returns an empty
  list for every metric.

  """

  def keys(self) -> Iterable[t.MetricKey]:
    return []

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    return {k: [] for k in ks}

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    return []


class LambdaReader(AbstractReader):
  """AbstractReader implementation that defers to a supplied lambda for its
  values. This allows you to escape the object-oriented programming paradigm,
  if you so choose.

  Args:
    read: Function called whenever reader.read(k) is called.
    read_all: Function called whenever reader.read_all(ks) is called.
    close: If supplied, this no-arg function will get called by this instance's
           `close` method.

  """

  def __init__(self,
               read: Optional[Callable[[t.MetricKey], List[t.Metric]]] = None,
               read_all: Optional[Callable[[List[t.MetricKey]],
                                           Dict[t.MetricKey,
                                                List[t.Metric]]]] = None,
               close: Optional[Callable[[], None]] = None):
    if read is None and read_all is None:
      raise ValueError(
          "Must supply one of `read` and `read_all` to `LambdaReader`.")

    self._read = read
    self._readall = read_all
    self._close = close

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    if self._readall is None:
      return super().read_all(ks)

    return self._readall(ks)

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    if self._read is None:
      return super().read(k)

    return self._read(k)

  def close(self) -> None:
    if self._close is not None:
      self._close()


class MemoryReader(AbstractReader, IterableReader):
  """Reader that queries the supplied dictionary for values.

  Args:
    m: Dictionary mapping metric keys to a list of accumulated metric values.

  """

  def __init__(self, m: Optional[Dict[str, List[t.Metric]]]):
    if m is None:
      m = {}

    self._m = m

  def keys(self) -> Iterable[t.Metric]:
    return self._m.keys()

  def read_all(self,
               ks: List[t.MetricKey]) -> Dict[t.MetricKey, List[t.Metric]]:
    return {str(k): self._m.get(str(k), []) for k in ks}

  def read(self, k: t.MetricKey) -> List[t.Metric]:
    return self._m.get(str(k), [])



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the SQL reader and reporter."""

import time
from contextlib import closing

import uv.sql.reporter as sr
import uv.sql.util as u

import pytest


def test_sqlite_file_exists(tmp_path):
  engine = u.sqlite_engine(str(tmp_path))

  # This is a temp URL, so nothing should exist at that location.
  assert not u.sqlite_file_exists(engine)
  assert not u.sqlite_file_exists(engine.url)

  # then we create the tables...
  sr.create_tables(engine)

  # now the database exists.
  assert u.sqlite_file_exists(engine)
  assert u.sqlite_file_exists(engine.url)

  # db suffix is automatically added:
  assert str(engine.url) == f"sqlite:///{tmp_path}.db"

  # If we explicitly add the suffix, the DB already exists.
  assert u.sqlite_file_exists(u.sqlite_engine(f"{tmp_path}.db"))


def test_rep_string():
  """Check that the custom representation string works for both of our types."""
  metric = sr.Metric(id=10, step=1)

  assert "id='10'" in str(metric)
  assert "step='1'" in str(metric)
  assert "value='None'" in str(metric)

  metric = sr.Experiment(id=10)
  assert "id='10'" in str(metric)
  assert "params='None'" in str(metric)


def test_of_sql_roundtrip(tmp_path):
  engine = u.sqlite_engine(str(tmp_path))

  # you can't make a reporter with an engine pointing to a nonexistent DB:
  with pytest.raises(Exception):
    sr.SQLReporter(engine, sr.Experiment(id=10), 0)

  # Same goes for reader.
  with pytest.raises(Exception):
    sr.SQLReader(engine, sr.Experiment(id=10), 0)

  sr.create_tables(engine)
  experiment = sr.new_experiment(engine, {"learning_rate": 0.01})

  with closing(sr.SQLReporter(engine, experiment, 0)) as reporter:
    with closing(reporter.reader()) as reader:
      with closing(sr.SQLReader(engine, experiment, 0)) as reader2:

        time.sleep(1)
        reporter.report_all(0, {"a": 1})
        reporter.report_all(1, {"a": 2, "b": 3})
        reporter.report_all(2, {"b": 4})

        a_entries = [{"step": 0, "value": 1.0}, {"step": 1, "value": 2.0}]
        b_entries = [{"step": 1, "value": 3.0}, {"step": 2, "value": 4.0}]

        assert list(reader.keys()) == ["a", "b"]
        assert reader.read("a") == a_entries
        assert reader.read("b") == b_entries

        assert reader.read_all(["a", "b"]) == {"a": a_entries, "b": b_entries}

        # Opening a reader directly has the same effect as calling reader() on
        # the reporter.
        assert reader.read_all(["a", "b"]) == reader2.read_all(["a", "b"])

        # Checking for a key that doesn't exist returns an empty list.
        assert reader.read("face") == []



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MLFlow reporter that conforms to UV's reporter interface."""

from typing import Dict

import mlflow
import uv.reporter.base as b
import uv.types as t


class MLFlowReporter(b.AbstractReporter):
  """Reporter implementation that logs metrics to mlflow."""

  def __init__(self):
    pass

  def report_param(self, k: str, v: str) -> None:
    mlflow.log_param(k, v)

  def report_params(self, m: Dict[str, str]) -> None:
    mlflow.log_params(m)

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    mlflow.log_metrics(m, step=step)

  def report(self, step: int, k: t.MetricKey, v: t.Metric) -> None:
    mlflow.log_metric(key=k, value=v, step=step)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reporter interface and implementations."""

import os
from contextlib import contextmanager
from typing import Dict

import mlflow as mlf
import uv.types as t
import uv.util.env as ue
from uv.reporter.base import AbstractReporter
from uv.reporter.store import NullReporter

_active_reporter = None


def get_reporter() -> AbstractReporter:
  """Returns the active reporter set using set_reporter()"""
  if _active_reporter is not None:
    return _active_reporter
  return NullReporter()


def set_reporter(r: AbstractReporter) -> AbstractReporter:
  """Set the globally available reporter instance. Returns its input."""
  global _active_reporter
  _active_reporter = r


@contextmanager
def active_reporter(r: AbstractReporter):
  old_reporter = _active_reporter
  globals()['_active_reporter'] = r
  yield
  globals()['_active_reporter'] = old_reporter


def report(step: int, k: t.MetricKey, v: t.Metric) -> None:
  """Accepts a step (an ordered int referencing some timestep), a metric key and
    a value, and persists the metric into the globally available reporter
    returned by uv.get_reporter().

  """
  return get_reporter().report(step, k, v)


def report_all(step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
  """Accepts a step (an ordered int referencing some timestep) and a dictionary
    of metric key => metric value, and persists the metric into the globally
    available reporter returned by uv.get_reporter().

  """
  return get_reporter().report_all(step, m)


def report_param(k: str, v: str) -> None:
  """Accepts a key and value parameter and logs these as parameters alongside the
    reported metrics. Reports to the globally available reporter returned by
    uv.get_reporter().

    """
  return get_reporter().report_param(k, v)


def report_params(m: Dict[str, str]) -> None:
  """Accepts a dict of parameter name -> value, and logs these as parameters
  alongside the reported metrics. Reports to the globally available reporter
  returned by uv.get_reporter().

  """
  return get_reporter().report_params(m)


def start_run(param_prefix: str = None, experiment_name: str = None, **args):
  """Close alias of mlflow.start_run. The only difference is that uv.start_run
  attempts to extract parameters from the environment and log those to the
  bound UV reporter using `report_params`.

  """
  if experiment_name is None:
    experiment_name = os.environ.get("MLFLOW_EXPERIMENT_NAME")

  # Make sure the experiment exists before the run starts.
  if experiment_name is not None:
    mlf.set_experiment(experiment_name)

  ret = mlf.start_run(**args)
  env_params = ue.extract_params(prefix=param_prefix)
  mlf.set_tags(env_params)
  return ret



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Code for manipulating and joining prefixes and suffixes to metric keys.

TODO convert attach into a single dispatch method, and make a new suffix
method.

"""

from collections import ChainMap
from typing import Dict, Iterable

import uv.types as t
import uv.util as u


def attach_s(s: t.MetricKey,
             a: t.Attachment,
             prefix: bool = True) -> t.MetricKey:
  """Attach the supplied prefix or suffix to the string s."""
  if not isinstance(s, str):
    s = str(s)

  if prefix:
    attached = u.wrap(a) + [s]
  else:
    attached = [s] + u.wrap(a)

  return ".".join(attached)


def attach_iter(xs: Iterable[t.MetricKey],
                a: t.Attachment,
                prefix: bool = True) -> Iterable[t.MetricKey]:
  """Attaches the supplied prefix or suffix to every item in the iterable."""
  return (attach_s(x, a, prefix) for x in xs)


def attach(m: Dict[t.MetricKey, t.Metric],
           a: t.Attachment,
           prefix: bool = True) -> Dict[t.MetricKey, t.Metric]:
  """Attaches the supplied prefix or suffix to every key in the dictionary."""
  return {attach_s(k, a, prefix): v for k, v in m.items()}


def _by_attachment(ms: Dict[str, Dict[t.MetricKey, t.Metric]],
                   prefix: bool = True) -> Dict[t.MetricKey, t.Metric]:
  """Collapse the prefixes into the key. If you do this, you can avoid using
  prefixed reporters.

  """
  return dict(ChainMap(*(attach(m, a, prefix) for a, m in ms.items())))


def by_prefix(
    ms: Dict[t.Attachment, Dict[t.MetricKey, t.Metric]]
) -> Dict[t.MetricKey, t.Metric]:
  """Collapse the prefixes into the key. If you do this, you can avoid using
  prefixed reporters.

  """
  return _by_attachment(ms, prefix=True)


def by_suffix(
    ms: Dict[t.Attachment, Dict[t.MetricKey, t.Metric]]
) -> Dict[t.MetricKey, t.Metric]:
  """Collapse the suffixes into the key. If you do this, you can avoid using
  suffixed reporters.

  """
  return _by_attachment(ms, prefix=False)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of the various reader store implementations."""

import hypothesis.strategies as st
from hypothesis import given

import pytest
import uv.reader.store as s
import uv.reporter.store as rs


@given(st.integers(min_value=0),
       st.dictionaries(st.text(min_size=1), st.integers()))
def test_memory_rt(step, m):
  """Check that any values pushed into a store are recovered."""
  mem = rs.MemoryReporter().stepped()
  reader = mem.reader()

  mem.report_all(step, m)

  expected = {k: [{"step": step, "value": v}] for k, v in m.items()}
  assert reader.read_all(m.keys()) == expected

  # The reader only knows about items from the input map.
  assert reader.keys() == m.keys()


@given(st.sets(st.text(min_size=1)))
def test_memory_empty(ks):
  """An empty reader should return nothing ever."""
  reader = s.MemoryReader({})
  assert reader.read_all(ks) == {k: [] for k in ks}

  # similar result if you don't supply ANY map.
  assert s.MemoryReader(m=None).read_all(ks) == {k: [] for k in ks}

  # same works for read implementation:
  for k in ks:
    assert reader.read(k) == []

  # no keys available from an empty reader.
  assert list(reader.keys()) == []


@given(st.text(min_size=1), st.lists(st.floats()))
def test_timeseries_all_ends_up_in_memory(k, vs):
  """Check that if many items are passed in for the same key, they'll all end up
  in memory in order.

  """
  mem = rs.MemoryReporter().stepped()
  reader = mem.reader()

  for i, v in enumerate(vs):
    mem.report(i, k, v)

  expected = [{"step": i, "value": v} for i, v in enumerate(vs)]
  assert reader.read(k) == expected


def test_memory_unit():
  """Non-hypothesis version to give a clear example."""
  mem = rs.MemoryReporter().stepped()
  reader = mem.reader()

  mem.report_all(0, {"a": 1, "b": 2})

  assert reader.read("a") == [{"step": 0, "value": 1}]
  assert reader.read("gunk") == []


@given(
    st.lists(st.dictionaries(st.text(min_size=1), st.integers(), max_size=100),
             max_size=10))
def test_null_empty(input_dicts):
  """An empty reader should return nothing ever."""

  null = rs.NullReporter()
  reader = null.reader()

  for i, m in enumerate(input_dicts):
    null.report_all(i, m)

    # let's report too, for good measure.
    for k, v in m.items():
      null.report(i, k, v)

  # no keys are available for reading.
  assert reader.keys() == []

  for m in input_dicts:
    ks = m.keys()
    assert reader.read_all(ks) == {k: [] for k in ks}

    # same for read:
    for k in ks:
      assert reader.read(k) == []


def test_null_unit():
  null = rs.NullReporter()
  reader = null.reader()

  null.report_all(0, {"1": "face"})
  null.report(0, "2", "cake")

  assert reader.read_all(["1", "2"]) == {"1": [], "2": []}
  assert reader.read("1") == []


@given(st.sets(st.text(min_size=1)))
def test_lambda_reader(ks):

  def _wrap_all(items):
    return {k: [k] for k in items}

  wrapped = _wrap_all(ks)

  # get a reader that returns a singleton value - just the requested key.
  reader = s.LambdaReader(read=lambda k: [k])
  ra = s.LambdaReader(read_all=_wrap_all)

  assert reader.read_all(ks) == wrapped
  assert ra.read_all(ks) == wrapped

  # Same test, but checking items one at a time.
  for k in ks:
    assert reader.read(k) == [k]
    assert ra.read(k) == [k]


def test_lambda_reader_errors():
  """The close function works, and you have to supply all required args."""

  # You have to supply a read or read_all fn.
  with pytest.raises(ValueError):
    read = s.LambdaReader()

  def explode():
    raise IOError("Don't close me!")

  read = s.LambdaReader(read=lambda k: [], close=explode)

  with pytest.raises(IOError):
    read.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uv.reporter.state as s
import uv.reporter.store as r


def test_global_reporter():
  r1 = r.MemoryReporter()
  reader1 = r1.reader()

  r2 = r.MemoryReporter()
  reader2 = r2.reader()

  r3 = r.MemoryReporter()
  reader3 = r3.reader()

  # By default, we get the null reporter and nothing happens.
  assert s.report(1, "face", 2) is None
  assert s.report_all(2, {"face": 2, "cake": 3}) is None

  s.set_reporter(r1)

  # boom!
  assert s.get_reporter() == r1

  # Now, reports go into the global reporter.
  s.report(1, "face", 2)
  s.report_all(2, {"face": 2, "cake": 3})

  # confirm:
  assert reader1.read("face") == [2, 2]
  assert reader1.read_all(["face", "cake"]) == {"face": [2, 2], "cake": [3]}

  # params are correctly reported.
  s.report_param("key_for_me", "value")
  s.report_params({"k": "v", "k2": "v2"})
  assert r1._params == {"key_for_me": "value", "k": "v", "k2": "v2"}

  with s.active_reporter(r2):

    # params get reported to the proper reporter.
    s.report_param("key", "value")
    assert r1._params == {"key_for_me": "value", "k": "v", "k2": "v2"}
    assert r2._params == {"key": "value"}

    # These report go through to r2, not r1.
    s.report(1, "hammer", 2)
    s.report_all(1, {"steve": 2, "john": 3})

    assert reader1.read_all(["hammer", "steve", "john"]) == {
        "hammer": [],
        "john": [],
        "steve": []
    }

    assert reader2.read_all(["hammer", "steve", "john"]) == {
        "hammer": [2],
        "john": [3],
        "steve": [2]
    }

    # these can nest:
    with s.active_reporter(r3):
      s.report(4, "deeper", 10)

      assert reader2.read("deeper") == []
      assert reader3.read("deeper") == [10]

    # when you leave a block, you hit the next reporter back.
    s.report(4, "deeper", 12)
    assert reader2.read("deeper") == [12]
    assert reader3.read("deeper") == [10]



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import find_packages, setup


def with_versioneer(f, default=None):
  """Attempts to execute the supplied single-arg function by passing it
versioneer if available; else, returns the default.

  """
  try:
    import versioneer
    return f(versioneer)
  except ModuleNotFoundError:
    return default


def readme():
  try:
    with open('README.md') as f:
      return f.read()
  except Exception:
    return None


CASFS_VERSION = "0.1.0"
REQUIRED_PACKAGES = [
    "casfs==0.1.1",
    "fs",
    "fs-gcsfs",
    "mlflow==1.10.0",
    "numpy>=1.18.0",
    "sqlalchemy",
    "tqdm>=4.42.1",
]

setup(
    name='uv-metrics',
    version=with_versioneer(lambda v: v.get_version()),
    cmdclass=with_versioneer(lambda v: v.get_cmdclass(), {}),
    description='Shared tooling for Blueshift research.',
    long_description=readme(),
    long_description_content_type="text/markdown",
    python_requires='>=3.6.0',
    author='Sam Ritchie',
    author_email='samritchie@google.com',
    url='https://github.com/google/uv-metrics',
    license='Apache-2.0',
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=REQUIRED_PACKAGES,
    extras_require={
        "tf": ["tensorflow"],
        "tf-gpu": ["tensorflow-gpu"],
    },
    include_package_data=True,
)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Types shared between readers and reporters."""

from typing import Any, Dict, Iterable, List, Tuple, Union

# Currently a metric can only be keyed by a string. Prefixes can't act as
# dictionary keys as they're not hashable. We could consider opening this up to
# tuples as well.
MetricKey = str
Metric = Any

Prefix = Union[MetricKey, List[str]]
Suffix = Union[MetricKey, List[str]]
Attachment = Union[Prefix, Suffix]



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for interacting with the system environment."""

import os
from typing import Dict, Optional

_ENV_VAR_PREFIX = "ENVVAR"


def extract_params(prefix: Optional[str] = None,
                   env: Optional[Dict[str, str]] = None) -> Dict[str, str]:
  """Some libraries that interact with UV pass experiment parameters via
  environment variables. This function returns a dict containing all
  environment keys (and their values) that begin with the supplied prefix and,
  optionally, an underscore.

  An environment like this:

  ENVVAR_MY_KEY=face
  ENVVAR_ANOTHER_KEY=sandwich
  ENVVARTHIRD_KEY=ham

  would return:

  {
    "my_key": "face",
    "another_key": "sandwich",
    "third_key": "ham"
  }

  """
  if prefix is None:
    prefix = _ENV_VAR_PREFIX

  if env is None:
    env = os.environ

  def relevant():
    for k, v in env.items():
      if k.startswith(prefix):
        stripped = k[len(prefix):].lstrip("_")
        yield stripped.lower(), v

  return dict(relevant())



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reporter implementation that relies on the PyFilesystem2 abstraction for
filesystems. See https://github.com/PyFilesystem/pyfilesystem2 for more
information on the library.

"""

from typing import Dict, Union

from fs.base import FS

import uv.fs.util as u
import uv.reader.base as rb
import uv.types as t
from uv.fs.reader import FSReader
from uv.reporter.base import AbstractReporter


class FSReporter(AbstractReporter):
  """AbstractReporter implementation backed by an instance of pyfilesystem2's FS
  abstraction. See https://docs.pyfilesystem.org/en/latest/ for more
  information.

  FSReporter serializes each metric as JSON

  TODO add a pluggable serializer, with a default jsonl implementation.

  Args:
    fs: Either an fs URI string, or an actual fs.base.FS object.

  """

  def __init__(self, fs: Union[FS, str]):
    self._fs = u.load_fs(fs)
    self._cache = u.HandleCache(self._fs)

  def _handle(self, k: t.MetricKey):
    """Returns an open handle that points to the file containing the metrics for
    k.

    """
    json_path = u.jsonl_path(k)
    return self._cache.open(json_path, mode='wb')

  def report_all(self, step: int, m: Dict[t.MetricKey, t.Metric]) -> None:
    for k, v in m.items():
      handle = self._handle(k)
      contents = u.jsonl_bytes(v)
      handle.write(contents)
      handle.flush()

  def reader(self) -> rb.AbstractReader:
    return FSReader(self._fs)

  def close(self) -> None:
    self._cache.close()
    self._fs.close()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of the various reporter store implementations."""

from collections import OrderedDict
import hypothesis.strategies as st
import numpy as np
from hypothesis import given

import pytest
import tests.uv.util.test_init as ti
import uv.reporter.store as rs


@given(st.dictionaries(st.text(min_size=1), st.integers()))
def test_lambda_reporter(m):

  # once we push all values into the store we expect metric => singleton.
  wrapped = {k: [v] for k, v in m.items()}
  double_wrapped = {k: [v, v] for k, v in m.items()}

  # make a memory reporter and a paired reader.
  mem = rs.MemoryReporter()
  reader = mem.reader()

  # these lambda reporters write to the backing memory store.
  r_reporter = rs.LambdaReporter(report=mem.report)
  ra_reporter = rs.LambdaReporter(report_all=mem.report_all)

  # report it ALL and check that everything made it in.
  ra_reporter.report_all(0, m)
  assert reader.read_all(m.keys()) == wrapped

  # now use ra_reporter's report method, which should delegate to report_all:
  for k, v in m.items():
    ra_reporter.report(0, k, v)

  # now we should have two copies for each key.
  assert reader.read_all(m.keys()) == double_wrapped

  # clear and confirm that everything is empty:
  mem.clear()
  assert reader.read_all(m.keys()) == {k: [] for k in m.keys()}

  # do the same thing again, but using the report interface this time.
  for k, v in m.items():
    r_reporter.report(0, k, v)

  assert reader.read_all(m.keys()) == wrapped

  # same thing as before, we check the report_all implementation when we only
  # supply a report function.
  r_reporter.report_all(0, m)
  assert reader.read_all(m.keys()) == double_wrapped


def test_lambda_reporter_errors():
  """The close function works, and you have to supply all required args."""

  # You have to supply a report or report_all fn.
  with pytest.raises(ValueError):
    rs.LambdaReporter()

  def explode():
    raise IOError("Don't close me!")

  report = rs.LambdaReporter(report=lambda _: None, close=explode)

  with pytest.raises(IOError):
    report.close()


def test_logging_reporter():
  """Check that the LoggingReporter actually logs out properly."""
  mem = ti.MemFile()

  reporter = rs.LoggingReporter(file=mem)

  # reporter handles non-native types like float32 just fine.
  reporter.report(0, "a", np.float32(1))

  # compound logging.
  m = OrderedDict([("a", 2), ("b", "cake")])
  reporter.report_all(1, m)

  # all items have been logged out.
  assert mem.items() == [
      'Step 0: a = 1.000', '\n', 'Step 1: a = 2.000, b = cake', '\n'
  ]


def test_logging_reporter_types():
  mem = ti.MemFile()
  reporter = rs.LoggingReporter(file=mem)

  v = np.array([1, 2, 3])
  reporter.report(0, "a", v)
  assert mem.items() == ['Step 0: a = {}'.format(str(v)), '\n']


def test_memory_reporter_types():
  # make a memory reporter and a paired reader.
  mem = rs.MemoryReporter()
  reader = mem.reader()

  # reporting to 1 and "1" should do the same thing.
  mem.report_all(0, {1: "face"})
  mem.report_all(1, {"1": "cake"})

  # int and string that stringify to the same key should return the values.
  assert reader.read(1) == ["face", "cake"]
  assert reader.read("1") == reader.read(1)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the filesystem reader and reporter."""

import string
import uuid
from contextlib import closing

import hypothesis.strategies as st
import uv.fs.util as u
from casfs import CASFS
from hypothesis import given
from uv.fs.reader import CASReader, FSReader
from uv.fs.reporter import FSReporter

import pytest


def test_fs_invalid():
  """Argument must be a string or a filesystem instance."""
  with pytest.raises(ValueError):
    FSReporter(100)

  with pytest.raises(ValueError):
    FSReporter(None)


def test_fs_roundtrip(tmpdir):
  dir_path = str(tmpdir)
  with closing(FSReporter(dir_path).stepped()) as reporter:
    with closing(reporter.reader()) as reader:
      with closing(FSReader(dir_path)) as reader2:

        reporter.report_all(0, {"a": 1})
        reporter.report_all(1, {"a": 2, "b": 3})
        reporter.report_all(2, {"b": 4})

        a_entries = [{"step": 0, "value": 1}, {"step": 1, "value": 2}]
        b_entries = [{"step": 1, "value": 3}, {"step": 2, "value": 4}]

        assert reader.read("a") == a_entries
        assert reader.read("b") == b_entries

        assert reader.read_all(["a", "b"]) == {"a": a_entries, "b": b_entries}

        # Opening a reader directly has the same effect as calling reader() on
        # the reporter.
        assert reader.read_all(["a", "b"]) == reader2.read_all(["a", "b"])

        # Checking for a key that doesn't exist returns an empty list.
        assert reader.read("face") == []


@given(
    st.dictionaries(st.text(alphabet=list(string.ascii_lowercase +
                                          string.digits),
                            min_size=1),
                    st.integers(),
                    max_size=100))
def test_fs_many(tmpdir, m):
  tmp = tmpdir.mkdir(str(uuid.uuid1()))
  dir_path = str(tmp)

  with closing(FSReporter(dir_path).stepped()) as reporter:
    with closing(reporter.reader()) as reader:
      reporter.report_all(0, m)

      # round-tripping keys
      expected = {}
      for k, v in m.items():
        expected.setdefault(k, []).append({"step": 0, "value": v})

      # check that the reader returns everything in the map written to the fs.
      assert reader.read_all(list(m.keys())) == expected

      # Check that the reader has all the goods!
      assert set(reader.keys()) == set(m.keys())


@pytest.fixture
def casfs(tmpdir):
  return u.get_cas(str(tmpdir))


def test_casfs_reader(tmpdir, casfs):
  """Check that we can persist metrics into a content-addressable store and get
  them back out.

  """
  dir_path = str(tmpdir)

  with closing(FSReporter(dir_path).stepped()) as reporter:
    with closing(reporter.reader()) as reader:
      reporter.report_all(0, {"a": 1})
      reporter.report_all(1, {"a": 2, "b": 3})
      reporter.report_all(2, {"b": 4})

      a_entries = [{"step": 0, "value": 1}, {"step": 1, "value": 2}]
      b_entries = [{"step": 1, "value": 3}, {"step": 2, "value": 4}]

      # Persist the whole metrics directory to the CAS. k is a key that
      # references the zipped metrics dir.
      #
      # If you store k.id (a string) in SQL, you can use it later to get access
      # to the metrics, as we'll see below.
      k = u.persist_to_cas_via_memory(casfs, dir_path)

      # This reader can read directly from the CasFS.
      with closing(CASReader(casfs, k.id)) as cas_reader:
        assert list(cas_reader.keys()) == list(reader.keys())

        # Contents are all identical:
        assert reader.read_all(["a", "b"]) == cas_reader.read_all(["a", "b"])

        # and the contents match what's expected:
        assert cas_reader.read_all(["a", "b"]) == {
            "a": a_entries,
            "b": b_entries
        }



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities grab bag."""

import datetime
import json
from functools import singledispatch
from typing import Any, List

import numpy as np
import tqdm


def to_metric(v: Any) -> float:
  """Converts the incoming item into something we can log.
  """
  if hasattr(v, 'numpy'):
    return v.numpy()

  return v


@singledispatch
def to_serializable(val):
  """Used by default."""

  if hasattr(val, 'numpy'):
    return val.numpy()

  return str(val)


@to_serializable.register(np.floating)
def ts_np_floating(val):
  """Used if *val* is an instance of numpy.floating."""
  return float(val)


@to_serializable.register(np.integer)
def ts_np_int(val):
  """Used if *val* is an instance of numpy.integer."""
  return int(val)


@to_serializable.register(np.ndarray)
def ts_np_array(val):
  """Convert a numpy array to a serializable list."""
  return val.tolist()


def json_str(item: Any) -> str:
  """Converts the supplied python object to a Python dictionary.

  The difference from json.dumps is that this method makes a best effort to
  serialize anything it finds within a nested structure. np.float32 instances,
  for example.

  """
  return json.dumps(item, default=to_serializable)


def uuid():  # pragma: no cover
  """Generates a sane UUID for use in test files and metric directories."""
  return datetime.datetime.now().strftime('%Y%m%d_%H%M%S')


def wrap(item: Any) -> List[Any]:
  """Ensures that the input is either a list, or wrapped in a list. Returns []
  for None.

  """
  if item is None:
    return []

  if isinstance(item, list):
    return item

  return [item]


def is_number(s):
  """Returns true if the supplied item can be converted into a float; false
  otherwise.

  """
  try:
    float(s)
    return True
  except Exception:
    return False


class TqdmFile():
  """Dummy file-like that will write to tqdm's 'write' method, which will in turn
  write back into the supplied file. Use this when you need to redirect some
  stream in such a way that won't cause the tqdm progress bar to reset.

  """
  file = None

  def __init__(self, file):
    self.file = file

  def write(self, x):
    if len(x.rstrip()) > 0:
      tqdm.tqdm.write(x, file=self.file, nolock=False)

  def flush(self):
    return getattr(self.file, "flush", lambda: None)()

  def isatty(self):
    return getattr(self.file, "isatty", lambda: False)()
